@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2008, 2009, 2010, 2011, 2012, 2013, 2015 Nicira, Inc.
+ * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -33,12 +34,7 @@ extern "C" {
  * headers to be aligned on a 4-byte boundary.  */
 enum { DP_NETDEV_HEADROOM = 2 + VLAN_HEADER_LEN };
 
-static inline void dp_packet_pad(struct dp_packet *p)
-{
-    if (dp_packet_size(p) < ETH_TOTAL_MIN) {
-        dp_packet_put_zeros(p, ETH_TOTAL_MIN - dp_packet_size(p));
-    }
-}
+enum { MAX_METERS = 5000 };
 
 bool dpif_is_netdev(const struct dpif *);
 
