@@ -27,6 +27,16 @@
 #include "tun-metadata.h"
 #include "openvswitch/nsh.h"
 
+static void
+set_eth(const struct eth_addr value_src,
+        struct eth_addr *value_dst,
+        struct eth_addr *mask_dst);
+
+static void
+set_eth_masked(const struct eth_addr value_src,
+               const struct eth_addr mask_src,
+               struct eth_addr *value_dst, struct eth_addr *mask_dst);
+
 /* Converts the flow in 'flow' into a match in 'match', with the given
  * 'wildcards'. */
 void
@@ -374,6 +384,78 @@ match_set_tun_erspan_hwid(struct match *match, uint8_t hwid)
     match_set_tun_erspan_hwid_masked(match, hwid, UINT8_MAX);
 }
 
+void
+match_set_tun_gtpu_flags_masked(struct match *match, uint8_t flags,
+                                uint8_t mask)
+{
+    match->wc.masks.tunnel.gtpu_flags = flags;
+    match->flow.tunnel.gtpu_flags = flags & mask;
+}
+
+void
+match_set_tun_gtpu_flags(struct match *match, uint8_t flags)
+{
+    match_set_tun_gtpu_flags_masked(match, flags, UINT8_MAX);
+}
+
+void
+match_set_tun_gtpu_msgtype_masked(struct match *match, uint8_t msgtype,
+                                  uint8_t mask)
+{
+    match->wc.masks.tunnel.gtpu_msgtype = msgtype;
+    match->flow.tunnel.gtpu_msgtype = msgtype & mask;
+}
+
+void
+match_set_tun_gtpu_msgtype(struct match *match, uint8_t msgtype)
+{
+    match_set_tun_gtpu_msgtype_masked(match, msgtype, UINT8_MAX);
+}
+
+void
+match_set_tun_eth_src(struct match *match, const struct eth_addr dl_src)
+{
+    set_eth(dl_src, &match->flow.tunnel.eth_src, &match->wc.masks.tunnel.eth_src);
+}
+
+void
+match_set_tun_eth_src_masked(struct match *match,
+                             const struct eth_addr dl_src,
+                             const struct eth_addr mask)
+{
+    set_eth_masked(dl_src, mask, &match->flow.tunnel.eth_src,
+                   &match->wc.masks.tunnel.eth_src);
+}
+
+void
+match_set_tun_eth_dst(struct match *match, const struct eth_addr dl_dst)
+{
+    set_eth(dl_dst, &match->flow.tunnel.eth_dst, &match->wc.masks.tunnel.eth_dst);
+}
+
+void
+match_set_tun_eth_dst_masked(struct match *match,
+                             const struct eth_addr dl_dst,
+                             const struct eth_addr mask)
+{
+    set_eth_masked(dl_dst, mask, &match->flow.tunnel.eth_dst,
+                   &match->wc.masks.tunnel.eth_dst);
+}
+
+void
+match_set_tun_type(struct match *match, uint8_t type)
+{
+    match->wc.masks.tunnel.type = UINT8_MAX;
+    match->flow.tunnel.type = type;
+}
+
+void
+match_set_tun_type_masked(struct match *match, uint8_t type, uint8_t mask)
+{
+    match->wc.masks.tunnel.type = mask;
+    match->flow.tunnel.type = type;
+}
+
 void
 match_set_in_port(struct match *match, ofp_port_t ofp_port)
 {
@@ -417,8 +499,14 @@ match_set_ct_state_masked(struct match *match, uint32_t ct_state, uint32_t mask)
 void
 match_set_ct_zone(struct match *match, uint16_t ct_zone)
 {
-    match->flow.ct_zone = ct_zone;
-    match->wc.masks.ct_zone = UINT16_MAX;
+    match_set_ct_zone_masked(match, ct_zone, UINT16_MAX);
+}
+
+void
+match_set_ct_zone_masked(struct match *match, uint16_t ct_zone, uint16_t mask)
+{
+    match->flow.ct_zone = ct_zone & mask;
+    match->wc.masks.ct_zone = mask;
 }
 
 void
@@ -893,6 +981,14 @@ match_set_nw_proto(struct match *match, uint8_t nw_proto)
     match->wc.masks.nw_proto = UINT8_MAX;
 }
 
+void
+match_set_nw_proto_masked(struct match *match,
+                          const uint8_t nw_proto, const uint8_t mask)
+{
+    match->flow.nw_proto = nw_proto;
+    match->wc.masks.nw_proto = mask;
+}
+
 void
 match_set_nw_src(struct match *match, ovs_be32 nw_src)
 {
@@ -986,6 +1082,30 @@ match_set_icmp_code(struct match *match, uint8_t icmp_code)
     match_set_tp_dst(match, htons(icmp_code));
 }
 
+void
+match_set_arp_opcode_masked(struct match *match,
+                            const uint8_t opcode,
+                            const uint8_t mask)
+{
+    match_set_nw_proto_masked(match, opcode, mask);
+}
+
+void
+match_set_arp_spa_masked(struct match *match,
+                         const ovs_be32 arp_spa,
+                         const ovs_be32 mask)
+{
+    match_set_nw_src_masked(match, arp_spa, mask);
+}
+
+void
+match_set_arp_tpa_masked(struct match *match,
+                         const ovs_be32 arp_tpa,
+                         const ovs_be32 mask)
+{
+    match_set_nw_dst_masked(match, arp_tpa, mask);
+}
+
 void
 match_set_arp_sha(struct match *match, const struct eth_addr sha)
 {
@@ -1080,6 +1200,19 @@ match_set_nd_target_masked(struct match *match,
     match->wc.masks.nd_target = *mask;
 }
 
+void
+match_set_nd_reserved (struct match *match, ovs_be32 value)
+{
+   match->flow.igmp_group_ip4 = value;
+   match->wc.masks.igmp_group_ip4 = OVS_BE32_MAX;
+}
+
+void
+match_set_nd_options_type(struct match *match, uint8_t option)
+{
+    match_set_tcp_flags(match, htons(option));
+}
+
 /* Returns true if 'a' and 'b' wildcard the same fields and have the same
  * values for fixed fields, otherwise false. */
 bool
@@ -1298,7 +1431,7 @@ format_flow_tunnel(struct ds *s, const struct match *match)
         ds_put_format(s, "tun_erspan_ver=%"PRIu8",", tnl->erspan_ver);
     }
     if (wc->masks.tunnel.erspan_idx && tnl->erspan_ver == 1) {
-       ds_put_format(s, "tun_erspan_idx=%#"PRIx32",", tnl->erspan_idx); 
+       ds_put_format(s, "tun_erspan_idx=%#"PRIx32",", tnl->erspan_idx);
     }
     if (wc->masks.tunnel.erspan_dir && tnl->erspan_ver == 2) {
         ds_put_format(s, "tun_erspan_dir=%"PRIu8",", tnl->erspan_dir);
@@ -1306,6 +1439,12 @@ format_flow_tunnel(struct ds *s, const struct match *match)
     if (wc->masks.tunnel.erspan_hwid && tnl->erspan_ver == 2) {
         ds_put_format(s, "tun_erspan_hwid=%#"PRIx8",", tnl->erspan_hwid);
     }
+    if (wc->masks.tunnel.gtpu_flags) {
+        ds_put_format(s, "gtpu_flags=%#"PRIx8",", tnl->gtpu_flags);
+    }
+    if (wc->masks.tunnel.gtpu_msgtype) {
+        ds_put_format(s, "gtpu_msgtype=%"PRIu8",", tnl->gtpu_msgtype);
+    }
     if (wc->masks.tunnel.flags & FLOW_TNL_F_MASK) {
         format_flags_masked(s, "tun_flags", flow_tun_flag_to_string,
                             tnl->flags & FLOW_TNL_F_MASK,
@@ -1313,6 +1452,12 @@ format_flow_tunnel(struct ds *s, const struct match *match)
                             FLOW_TNL_F_MASK);
         ds_put_char(s, ',');
     }
+    format_eth_masked(s, "tun_eth_src", tnl->eth_src, wc->masks.tunnel.eth_src);
+    format_eth_masked(s, "tun_eth_dst", tnl->eth_dst, wc->masks.tunnel.eth_dst);
+    if (wc->masks.tunnel.type) {
+        ds_put_format(s, "tun_type=%s,",
+                      tun_type_type2str((enum tun_type) tnl->type));
+    }
     tun_metadata_match_format(s, match);
 }
 
@@ -1377,7 +1522,7 @@ match_format(const struct match *match,
     bool is_megaflow = false;
     int i;
 
-    BUILD_ASSERT_DECL(FLOW_WC_SEQ == 41);
+    BUILD_ASSERT_DECL(FLOW_WC_SEQ == 42);
 
     if (priority != OFP_DEFAULT_PRIORITY) {
         ds_put_format(s, "%spriority=%s%d,",
@@ -1471,8 +1616,6 @@ match_format(const struct match *match,
                 skip_proto = true;
                 if (f->nw_proto == IPPROTO_ICMP) {
                     ds_put_format(s, "%sicmp%s,", colors.value, colors.end);
-                } else if (f->nw_proto == IPPROTO_IGMP) {
-                    ds_put_format(s, "%sigmp%s,", colors.value, colors.end);
                 } else if (f->nw_proto == IPPROTO_TCP) {
                     ds_put_format(s, "%stcp%s,", colors.value, colors.end);
                 } else if (f->nw_proto == IPPROTO_UDP) {
@@ -1654,7 +1797,7 @@ match_format(const struct match *match,
     format_be32_masked(s, "mpls_lse1", f->mpls_lse[1], wc->masks.mpls_lse[1]);
     format_be32_masked(s, "mpls_lse2", f->mpls_lse[2], wc->masks.mpls_lse[2]);
 
-    switch (wc->masks.nw_frag) {
+    switch (wc->masks.nw_frag & FLOW_NW_FRAG_MASK) {
     case FLOW_NW_FRAG_ANY | FLOW_NW_FRAG_LATER:
         ds_put_format(s, "%snw_frag=%s%s,", colors.param, colors.end,
                       f->nw_frag & FLOW_NW_FRAG_ANY
@@ -1676,10 +1819,6 @@ match_format(const struct match *match,
         f->nw_proto == IPPROTO_ICMP) {
         format_be16_masked(s, "icmp_type", f->tp_src, wc->masks.tp_src);
         format_be16_masked(s, "icmp_code", f->tp_dst, wc->masks.tp_dst);
-    } else if (dl_type == htons(ETH_TYPE_IP) &&
-               f->nw_proto == IPPROTO_IGMP) {
-        format_be16_masked(s, "igmp_type", f->tp_src, wc->masks.tp_src);
-        format_be16_masked(s, "igmp_code", f->tp_dst, wc->masks.tp_dst);
     } else if (dl_type == htons(ETH_TYPE_IPV6) &&
                f->nw_proto == IPPROTO_ICMPV6) {
         format_be16_masked(s, "icmp_type", f->tp_src, wc->masks.tp_src);
@@ -1688,6 +1827,14 @@ match_format(const struct match *match,
                             &wc->masks.nd_target);
         format_eth_masked(s, "nd_sll", f->arp_sha, wc->masks.arp_sha);
         format_eth_masked(s, "nd_tll", f->arp_tha, wc->masks.arp_tha);
+        if (wc->masks.igmp_group_ip4) {
+            format_be32_masked(s,"nd_reserved", f->igmp_group_ip4,
+                               wc->masks.igmp_group_ip4);
+        }
+        if (wc->masks.tcp_flags) {
+            format_be16_masked(s,"nd_options_type", f->tcp_flags,
+                               wc->masks.tcp_flags);
+        }
     } else {
         format_be16_masked(s, "tp_src", f->tp_src, wc->masks.tp_src);
         format_be16_masked(s, "tp_dst", f->tp_dst, wc->masks.tp_dst);
