@@ -55,15 +55,37 @@ nl_msg_genlmsghdr(const struct ofpbuf *msg)
     return ofpbuf_at(msg, NLMSG_HDRLEN, GENL_HDRLEN);
 }
 
+/* Parses the ext ack netlink attributes and, if successful, a pointer
+ * to the error message, if included, is stored in '*errmsg'. */
+static void
+nl_parse_ext_ack(const struct ofpbuf *msg, size_t offset, const char **errmsg)
+{
+    static const struct nl_policy policy[] = {
+        [NLMSGERR_ATTR_MSG]  = { .type = NL_A_STRING, .optional = true },
+    };
+    struct nlattr *attrs[ARRAY_SIZE(policy)];
+
+    if (!nl_policy_parse(msg, offset, policy, attrs, ARRAY_SIZE(policy))) {
+        VLOG_ERR_RL(&rl, "Failed to parse extended ack data");
+        return;
+    }
+
+    if (attrs[NLMSGERR_ATTR_MSG]) {
+        *errmsg = nl_attr_get_string(attrs[NLMSGERR_ATTR_MSG]);
+    }
+}
+
 /* If 'buffer' is a NLMSG_ERROR message, stores 0 in '*errorp' if it is an ACK
  * message, otherwise a positive errno value, and returns true.  If 'buffer' is
  * not an NLMSG_ERROR message, returns false.
  *
  * 'msg' must be at least as large as a nlmsghdr. */
 bool
-nl_msg_nlmsgerr(const struct ofpbuf *msg, int *errorp)
+nl_msg_nlmsgerr(const struct ofpbuf *msg, int *errorp, const char **attr_msg)
 {
-    if (nl_msg_nlmsghdr(msg)->nlmsg_type == NLMSG_ERROR) {
+    struct nlmsghdr *nlh = nl_msg_nlmsghdr(msg);
+
+    if (nlh->nlmsg_type == NLMSG_ERROR) {
         struct nlmsgerr *err = ofpbuf_at(msg, NLMSG_HDRLEN, sizeof *err);
         int code = EPROTO;
         if (!err) {
@@ -71,6 +93,15 @@ nl_msg_nlmsgerr(const struct ofpbuf *msg, int *errorp)
                         msg->size, NLMSG_HDRLEN + sizeof *err);
         } else if (err->error <= 0 && err->error > INT_MIN) {
             code = -err->error;
+            if (attr_msg && err->error != 0 &&
+                (nlh->nlmsg_flags & NLM_F_ACK_TLVS)) {
+                size_t offt =  NLMSG_HDRLEN + sizeof *err;
+
+                if (!(nlh->nlmsg_flags & NLM_F_CAPPED)) {
+                    offt += (err->msg.nlmsg_len - NLMSG_HDRLEN);
+                }
+                nl_parse_ext_ack(msg, offt, attr_msg);
+            }
         }
         if (errorp) {
             *errorp = code;
@@ -169,6 +200,14 @@ nl_msg_put(struct ofpbuf *msg, const void *data, size_t size)
     memcpy(nl_msg_put_uninit(msg, size), data, size);
 }
 
+static void
+nl_msg_zero_pad(char *pad_ptr, size_t pad_size)
+{
+    while (pad_size--) {
+        *pad_ptr++ = 0;
+    }
+}
+
 /* Appends 'size' bytes of data, plus Netlink padding if needed, to the tail
  * end of 'msg', reallocating and copying its data if necessary.  Returns a
  * pointer to the first byte of the new data, which is left uninitialized. */
@@ -177,9 +216,7 @@ nl_msg_put_uninit(struct ofpbuf *msg, size_t size)
 {
     size_t pad = PAD_SIZE(size, NLMSG_ALIGNTO);
     char *p = ofpbuf_put_uninit(msg, size + pad);
-    if (pad) {
-        memset(p + size, 0, pad);
-    }
+    nl_msg_zero_pad(p + size, pad);
     return p;
 }
 
@@ -200,9 +237,7 @@ nl_msg_push_uninit(struct ofpbuf *msg, size_t size)
 {
     size_t pad = PAD_SIZE(size, NLMSG_ALIGNTO);
     char *p = ofpbuf_push_uninit(msg, size + pad);
-    if (pad) {
-        memset(p + size, 0, pad);
-    }
+    nl_msg_zero_pad(p + size, pad);
     return p;
 }
 
@@ -235,13 +270,15 @@ nl_msg_put_unspec_zero(struct ofpbuf *msg, uint16_t type, size_t size)
 
 /* Appends a Netlink attribute of the given 'type' and the 'size' bytes of
  * 'data' as its payload, to the tail end of 'msg', reallocating and copying
- * its data if necessary.  Returns a pointer to the first byte of data in the
- * attribute, which is left uninitialized. */
+ * its data if necessary. */
 void
 nl_msg_put_unspec(struct ofpbuf *msg, uint16_t type,
                   const void *data, size_t size)
 {
-    memcpy(nl_msg_put_unspec_uninit(msg, type, size), data, size);
+    void *ptr;
+
+    ptr = nl_msg_put_unspec_uninit(msg, type, size);
+    nullable_memcpy(ptr, data, size);
 }
 
 /* Appends a Netlink attribute of the given 'type' and no payload to 'msg'.
@@ -486,7 +523,7 @@ size_t
 nl_msg_start_nested(struct ofpbuf *msg, uint16_t type)
 {
     size_t offset = msg->size;
-    nl_msg_put_unspec(msg, type, NULL, 0);
+    nl_msg_put_unspec_uninit(msg, type | NLA_F_NESTED, 0);
     return offset;
 }
 
@@ -496,9 +533,36 @@ void
 nl_msg_end_nested(struct ofpbuf *msg, size_t offset)
 {
     struct nlattr *attr = ofpbuf_at_assert(msg, offset, sizeof *attr);
+    ovs_assert(!nl_attr_oversized(msg->size - offset - NLA_HDRLEN));
     attr->nla_len = msg->size - offset;
 }
 
+/* Cancel a nested Netlink attribute in 'msg'.  'offset' should be the value
+ * returned by nl_msg_start_nested(). */
+void
+nl_msg_cancel_nested(struct ofpbuf *msg, size_t offset)
+{
+    msg->size = offset;
+}
+
+/* Same as nls_msg_end_nested() when the nested Netlink contains non empty
+ * message. Otherwise, drop the nested message header from 'msg'.
+ *
+ * Return true if the nested message has been dropped.  */
+bool
+nl_msg_end_non_empty_nested(struct ofpbuf *msg, size_t offset)
+{
+    nl_msg_end_nested(msg, offset);
+
+    struct nlattr *attr = ofpbuf_at_assert(msg, offset, sizeof *attr);
+    if (!nl_attr_get_size(attr)) {
+        nl_msg_cancel_nested(msg, offset);
+        return true;
+    } else {
+        return false;
+    }
+}
+
 /* Appends a nested Netlink attribute of the given 'type', with the 'size'
  * bytes of content starting at 'data', to 'msg'. */
 void
@@ -712,6 +776,20 @@ nl_attr_get_nested(const struct nlattr *nla, struct ofpbuf *nested)
     ofpbuf_use_const(nested, nl_attr_get(nla), nl_attr_get_size(nla));
 }
 
+/* Returns the Ethernet Address value in 'nla''s payload. */
+struct eth_addr
+nl_attr_get_eth_addr(const struct nlattr *nla)
+{
+    return NL_ATTR_GET_AS(nla, struct eth_addr);
+}
+
+/* Returns the Infiniband LL Address value in 'nla''s payload. */
+struct ib_addr
+nl_attr_get_ib_addr(const struct nlattr *nla)
+{
+    return NL_ATTR_GET_AS(nla, struct ib_addr);
+}
+
 /* Default minimum payload size for each type of attribute. */
 static size_t
 min_attr_len(enum nl_attr_type type)
@@ -728,6 +806,7 @@ min_attr_len(enum nl_attr_type type)
     case NL_A_FLAG: return 0;
     case NL_A_IPV6: return 16;
     case NL_A_NESTED: return 0;
+    case NL_A_LL_ADDR: return 6; /* ETH_ALEN */
     case N_NL_ATTR_TYPES: default: OVS_NOT_REACHED();
     }
 }
@@ -748,6 +827,7 @@ max_attr_len(enum nl_attr_type type)
     case NL_A_FLAG: return SIZE_MAX;
     case NL_A_IPV6: return 16;
     case NL_A_NESTED: return SIZE_MAX;
+    case NL_A_LL_ADDR: return 20; /* INFINIBAND_ALEN */
     case N_NL_ATTR_TYPES: default: OVS_NOT_REACHED();
     }
 }
