@@ -13,8 +13,6 @@
  * limitations under the License. */
 
 #include <config.h>
-#include "tunnel.h"
-
 #include <errno.h>
 
 #include "byte-order.h"
@@ -37,6 +35,7 @@
 #include "tunnel.h"
 #include "openvswitch/vlog.h"
 #include "unaligned.h"
+#include "unixctl.h"
 #include "ofproto-dpif.h"
 #include "netdev-vport.h"
 
@@ -121,7 +120,10 @@ static struct tnl_port *tnl_find_ofport(const struct ofport_dpif *)
 
 static uint32_t tnl_hash(struct tnl_match *);
 static void tnl_match_fmt(const struct tnl_match *, struct ds *);
-static char *tnl_port_fmt(const struct tnl_port *) OVS_REQ_RDLOCK(rwlock);
+static char *tnl_port_to_string(const struct tnl_port *)
+    OVS_REQ_RDLOCK(rwlock);
+static void tnl_port_format(const struct tnl_port *, struct ds *)
+    OVS_REQ_RDLOCK(rwlock);
 static void tnl_port_mod_log(const struct tnl_port *, const char *action)
     OVS_REQ_RDLOCK(rwlock);
 static const char *tnl_port_get_name(const struct tnl_port *)
@@ -129,6 +131,8 @@ static const char *tnl_port_get_name(const struct tnl_port *)
 static void tnl_port_del__(const struct ofport_dpif *, odp_port_t)
     OVS_REQ_WRLOCK(rwlock);
 
+static unixctl_cb_func tnl_unixctl_list;
+
 void
 ofproto_tunnel_init(void)
 {
@@ -136,6 +140,8 @@ ofproto_tunnel_init(void)
 
     if (ovsthread_once_start(&once)) {
         fat_rwlock_init(&rwlock);
+        unixctl_command_register("ofproto/list-tunnels", "", 0, 0,
+                                 tnl_unixctl_list, NULL);
         ovsthread_once_done(&once);
     }
 }
@@ -301,7 +307,6 @@ tnl_port_del(const struct ofport_dpif *ofport, odp_port_t odp_port)
 const struct ofport_dpif *
 tnl_port_receive(const struct flow *flow) OVS_EXCLUDED(rwlock)
 {
-    char *pre_flow_str = NULL;
     const struct ofport_dpif *ofport;
     struct tnl_port *tnl_port;
 
@@ -309,28 +314,20 @@ tnl_port_receive(const struct flow *flow) OVS_EXCLUDED(rwlock)
     tnl_port = tnl_find(flow);
     ofport = tnl_port ? tnl_port->ofport : NULL;
     if (!tnl_port) {
-        char *flow_str = flow_to_string(flow, NULL);
-
-        VLOG_WARN_RL(&rl, "receive tunnel port not found (%s)", flow_str);
-        free(flow_str);
+        if (!VLOG_DROP_WARN(&rl)) {
+            char *flow_str = flow_to_string(flow, NULL);
+            VLOG_WARN("receive tunnel port not found (%s)", flow_str);
+            free(flow_str);
+        }
         goto out;
     }
 
     if (!VLOG_DROP_DBG(&dbg_rl)) {
-        pre_flow_str = flow_to_string(flow, NULL);
-    }
-
-    if (pre_flow_str) {
-        char *post_flow_str = flow_to_string(flow, NULL);
-        char *tnl_str = tnl_port_fmt(tnl_port);
-        VLOG_DBG("flow received\n"
-                 "%s"
-                 " pre: %s\n"
-                 "post: %s",
-                 tnl_str, pre_flow_str, post_flow_str);
+        char *flow_str = flow_to_string(flow, NULL);
+        char *tnl_str = tnl_port_to_string(tnl_port);
+        VLOG_DBG("tunnel port %s receive from flow %s", tnl_str, flow_str);
         free(tnl_str);
-        free(pre_flow_str);
-        free(post_flow_str);
+        free(flow_str);
     }
 
 out:
@@ -367,17 +364,29 @@ void
 tnl_wc_init(struct flow *flow, struct flow_wildcards *wc)
 {
     if (tnl_port_should_receive(flow)) {
-        wc->masks.tunnel.tun_id = OVS_BE64_MAX;
-        if (flow->tunnel.ip_dst) {
-            wc->masks.tunnel.ip_src = OVS_BE32_MAX;
-            wc->masks.tunnel.ip_dst = OVS_BE32_MAX;
-        } else {
-            wc->masks.tunnel.ipv6_src = in6addr_exact;
-            wc->masks.tunnel.ipv6_dst = in6addr_exact;
+        if (!(flow->tunnel.flags & FLOW_TNL_F_EXPLICIT)) {
+            wc->masks.tunnel.tun_id = OVS_BE64_MAX;
+            if (flow->tunnel.ip_dst) {
+                wc->masks.tunnel.ip_src = OVS_BE32_MAX;
+                wc->masks.tunnel.ip_dst = OVS_BE32_MAX;
+            } else {
+                wc->masks.tunnel.ipv6_src = in6addr_exact;
+                wc->masks.tunnel.ipv6_dst = in6addr_exact;
+            }
         }
+
+        /* OpenFlow does not support the +csum flag on tunnels and
+         * netdev offloads ignore that match. A packet in SW datapath
+         * however would generate different dp-flows depending on
+         * whether its csum is set, breaking some offload implementation.
+         * Align all layers to ignore the csum in all matches, even
+         * in SW. */
         wc->masks.tunnel.flags = (FLOW_TNL_F_DONT_FRAGMENT |
-                                  FLOW_TNL_F_CSUM |
+                                  /* FLOW_TNL_F_CSUM - See comment above */
                                   FLOW_TNL_F_KEY);
+        if (flow->tunnel.flags & FLOW_TNL_F_EXPLICIT) {
+            wc->masks.tunnel.flags |= FLOW_TNL_F_EXPLICIT;
+        }
         wc->masks.tunnel.ip_tos = UINT8_MAX;
         wc->masks.tunnel.ip_ttl = 0;
         /* The tp_src and tp_dst members in flow_tnl are set to be always
@@ -419,7 +428,7 @@ tnl_port_send(const struct ofport_dpif *ofport, struct flow *flow,
         pre_flow_str = flow_to_string(flow, NULL);
     }
 
-    if (!cfg->ip_src_flow) {
+    if (!cfg->is_explicit && !cfg->ip_src_flow) {
         flow->tunnel.ip_src = in6_addr_get_mapped_ipv4(&tnl_port->match.ipv6_src);
         if (!flow->tunnel.ip_src) {
             flow->tunnel.ipv6_src = tnl_port->match.ipv6_src;
@@ -427,7 +436,7 @@ tnl_port_send(const struct ofport_dpif *ofport, struct flow *flow,
             flow->tunnel.ipv6_src = in6addr_any;
         }
     }
-    if (!cfg->ip_dst_flow) {
+    if (!cfg->is_explicit && !cfg->ip_dst_flow) {
         flow->tunnel.ip_dst = in6_addr_get_mapped_ipv4(&tnl_port->match.ipv6_dst);
         if (!flow->tunnel.ip_dst) {
             flow->tunnel.ipv6_dst = tnl_port->match.ipv6_dst;
@@ -435,9 +444,11 @@ tnl_port_send(const struct ofport_dpif *ofport, struct flow *flow,
             flow->tunnel.ipv6_dst = in6addr_any;
         }
     }
-    flow->tunnel.tp_dst = cfg->dst_port;
-    if (!cfg->out_key_flow) {
-        flow->tunnel.tun_id = cfg->out_key;
+    if (!cfg->is_explicit) {
+        flow->tunnel.tp_dst = cfg->dst_port;
+        if (!cfg->out_key_flow) {
+            flow->tunnel.tun_id = cfg->out_key;
+        }
     }
 
     if (cfg->ttl_inherit && is_ip_any(flow)) {
@@ -468,16 +479,32 @@ tnl_port_send(const struct ofport_dpif *ofport, struct flow *flow,
     flow->tunnel.flags &= ~(FLOW_TNL_F_MASK & ~FLOW_TNL_PUB_F_MASK);
     flow->tunnel.flags |= (cfg->dont_fragment ? FLOW_TNL_F_DONT_FRAGMENT : 0)
         | (cfg->csum ? FLOW_TNL_F_CSUM : 0)
-        | (cfg->out_key_present ? FLOW_TNL_F_KEY : 0);
+        | ((cfg->out_key_present | cfg->is_explicit) ? FLOW_TNL_F_KEY : 0);
 
     if (cfg->set_egress_pkt_mark) {
         flow->pkt_mark = cfg->egress_pkt_mark;
         wc->masks.pkt_mark = UINT32_MAX;
     }
 
+    if (!cfg->erspan_ver_flow) {
+        flow->tunnel.erspan_ver = cfg->erspan_ver;
+    }
+
+    if (!cfg->erspan_idx_flow) {
+        flow->tunnel.erspan_idx = cfg->erspan_idx;
+    }
+
+    if (!cfg->erspan_dir_flow) {
+        flow->tunnel.erspan_dir = cfg->erspan_dir;
+    }
+
+    if (!cfg->erspan_hwid_flow) {
+        flow->tunnel.erspan_hwid = cfg->erspan_hwid;
+    }
+
     if (pre_flow_str) {
         char *post_flow_str = flow_to_string(flow, NULL);
-        char *tnl_str = tnl_port_fmt(tnl_port);
+        char *tnl_str = tnl_port_to_string(tnl_port);
         VLOG_DBG("flow sent\n"
                  "%s"
                  " pre: %s\n"
@@ -652,53 +679,58 @@ tnl_port_mod_log(const struct tnl_port *tnl_port, const char *action)
     }
 }
 
-static char *
-tnl_port_fmt(const struct tnl_port *tnl_port) OVS_REQ_RDLOCK(rwlock)
+static void OVS_REQ_RDLOCK(rwlock)
+tnl_port_format(const struct tnl_port *tnl_port, struct ds *ds)
 {
     const struct netdev_tunnel_config *cfg =
         netdev_get_tunnel_config(tnl_port->netdev);
-    struct ds ds = DS_EMPTY_INITIALIZER;
 
-    ds_put_format(&ds, "port %"PRIu32": %s (%s: ", tnl_port->match.odp_port,
+    ds_put_format(ds, "port %"PRIu32": %s (%s: ", tnl_port->match.odp_port,
                   tnl_port_get_name(tnl_port),
                   netdev_get_type(tnl_port->netdev));
-    tnl_match_fmt(&tnl_port->match, &ds);
+    tnl_match_fmt(&tnl_port->match, ds);
 
     if (cfg->out_key != cfg->in_key ||
         cfg->out_key_present != cfg->in_key_present ||
         cfg->out_key_flow != cfg->in_key_flow) {
-        ds_put_cstr(&ds, ", out_key=");
+        ds_put_cstr(ds, ", out_key=");
         if (!cfg->out_key_present) {
-            ds_put_cstr(&ds, "none");
+            ds_put_cstr(ds, "none");
         } else if (cfg->out_key_flow) {
-            ds_put_cstr(&ds, "flow");
+            ds_put_cstr(ds, "flow");
         } else {
-            ds_put_format(&ds, "%#"PRIx64, ntohll(cfg->out_key));
+            ds_put_format(ds, "%#"PRIx64, ntohll(cfg->out_key));
         }
     }
 
     if (cfg->ttl_inherit) {
-        ds_put_cstr(&ds, ", ttl=inherit");
+        ds_put_cstr(ds, ", ttl=inherit");
     } else {
-        ds_put_format(&ds, ", ttl=%"PRIu8, cfg->ttl);
+        ds_put_format(ds, ", ttl=%"PRIu8, cfg->ttl);
     }
 
     if (cfg->tos_inherit) {
-        ds_put_cstr(&ds, ", tos=inherit");
+        ds_put_cstr(ds, ", tos=inherit");
     } else if (cfg->tos) {
-        ds_put_format(&ds, ", tos=%#"PRIx8, cfg->tos);
+        ds_put_format(ds, ", tos=%#"PRIx8, cfg->tos);
     }
 
     if (!cfg->dont_fragment) {
-        ds_put_cstr(&ds, ", df=false");
+        ds_put_cstr(ds, ", df=false");
     }
 
     if (cfg->csum) {
-        ds_put_cstr(&ds, ", csum=true");
+        ds_put_cstr(ds, ", csum=true");
     }
 
-    ds_put_cstr(&ds, ")\n");
+    ds_put_cstr(ds, ")\n");
+}
 
+static char * OVS_REQ_RDLOCK(rwlock)
+tnl_port_to_string(const struct tnl_port *tnl_port)
+{
+    struct ds ds = DS_EMPTY_INITIALIZER;
+    tnl_port_format(tnl_port, &ds);
     return ds_steal_cstr(&ds);
 }
 
@@ -734,3 +766,26 @@ tnl_port_build_header(const struct ofport_dpif *ofport,
 
     return res;
 }
+
+static void
+tnl_unixctl_list(struct unixctl_conn *conn,
+                 int argc OVS_UNUSED, const char *argv[] OVS_UNUSED,
+                 void *aux OVS_UNUSED)
+{
+    struct ds reply = DS_EMPTY_INITIALIZER;
+
+    fat_rwlock_rdlock(&rwlock);
+    for (int i = 0; i < N_MATCH_TYPES; i++) {
+        struct hmap *map = tnl_match_maps[i];
+        if (map) {
+            struct tnl_port *tnl_port;
+            HMAP_FOR_EACH (tnl_port, match_node, map) {
+                tnl_port_format(tnl_port, &reply);
+            }
+        }
+    }
+    fat_rwlock_unlock(&rwlock);
+
+    unixctl_command_reply(conn, ds_cstr(&reply));
+    ds_destroy(&reply);
+}
